import numpy as np
import scipy as sp
from scipy.integrate import ode,simps,quad
from scipy.fftpack import fft, ifft, fftfreq
from scipy.signal import fftconvolve
import matplotlib.pylab  as plt

#from tempfile import TemporaryFile

import math as ma




Temp = 200. # Temperature of thermal bath
t1 = 0.0007  #final time
dt = 0.00001 #time step
freq_cut= 100.  # frequency cutoff
 

hx0=[0.,1.] # initial condition for h
Hx0=[1.,0.] # initial condition for H
t0=0. # initial time
tf_ind =2 # initial value for the final time index for Decay factor
omega=1 #omega


s_w = 10.
x_w = 5.

def fGRw(w): 
    # Retarded Green function 
     return  -(0.+1.j)*(w**3+(2*np.pi*Temp)**2*w)*freq_cut**2/(w**2+freq_cut**2) 
     
# static string case : -(0.+1.j)*(w**3+(2*np.pi*Temp)**2*w)*freq_cut**2/(w**2+freq_cut**2)     
#ohmic case:  -(0.+1.j)*w*freq_cut**2/(w**2+freq_cut**2) 
 
def fG_sym(w):
    return -fGRw(w).imag/sp.tanh(w/2/Temp)



t_leng = int(t1/dt)
full_t_ind = np.arange(t_leng)
full_t_vec = np.linspace(0,t1,t_leng)
full_w_vec = np.linspace(-freq_cut, freq_cut, t_leng)
#full_w_vec = np.linspace(-1/dt, 1/dt, t_leng)



# Fourier Transform of g(t): int_0^t g(tau) exp(-i*w*tau) dtau
def gFT(gxall, t_all, gt_ind, gw_vec, gw_ind):
    g_ftw = []  
    for i in gw_ind:
        g_integ = np.array([gxall[k]*np.exp((0.-1.j)*gw_vec[i]*t_all[k])/np.sqrt(2*np.pi) for k in gt_ind])
        g_ftw.append(simps(g_integ, t_all))
    return g_ftw
  
    
# Inverse Fourier Transform of g(w): int_0^Omega g(w) exp(-i*w*t) dt       
def ginvFT(gwall, w_all, gw_ind, gt_vec, gt_ind):
    g_iftw = []  
    for i in gw_ind:
        g_integ = np.array([gwall[k]*np.exp((0.+1.j)*gt_vec[i]*w_all[k])/np.sqrt(2*np.pi) for k in gw_ind])
        g_iftw.append(simps(g_integ, w_all))
    return g_iftw
                
                 


def h(t,y,convo):
    # RHS of the coupled first-order ODE
    hx0=y[0] # x
    hx1=y[1] # x'
    return [hx1,-omega*omega*hx0-convo]
    
def H(t,y,convo):
    # RHS of the coupled first-order ODE
    Hx0=y[0] # x
    Hx1=y[1] # x'
    return [Hx1,-omega*omega*Hx0-convo]
          
  
    
def convoluh(t,G_vec):
    # Convolution Integral
    Nmax=int(t/dt)
    if Nmax != 0:
        ht_series=htall[0:Nmax]
        hk_series=hxall[0:Nmax]
        for i in range(Nmax):
            tau=i*dt
            hk_series[i]*=G_vec[Nmax-i]
        return simps(hk_series, ht_series)
    else:
        return 0.0
        


        
# Define a00, a01 and a11 for a given t_f         
def aij(ag0_ftw, ag1_ftw, w_vec, Gsym0, w_ind):
     Gsymw = np.array([Gsym0[i] for i in w_ind])   # symmetric Green function
     ax00 = 2*simps(ag0_ftw*Gsymw*np.conjugate(ag0_ftw), w_vec).real  # a00 
     ax01 = 2*simps(ag0_ftw*Gsymw*np.conjugate(ag1_ftw), w_vec).real  # a01
     ax11 = 2*simps(ag1_ftw*Gsymw*np.conjugate(ag1_ftw), w_vec).real  # a11 
     return ax00, ax01, ax11


         

def Wigner_Renyi(sw,xw,ht,f0v0,f1v0,f0vt,f1vt,a00,a01,a11):
     A =1/(2*np.pi*ht)             
     C = (sw*f0v0)**2/4 +1/4/sw**2 + a00/2 
     N_inv = 2*np.sqrt(np.pi)*sw*(1+np.exp(-(xw/sw)**2))
     Ls2_inv = np.abs(f1v0**2/4/C)
     Ld2_inv = np.abs((sw*f0vt)**2/4 + a11/2 - (sw**2*f0vt*f0v0/2-a01)**2/4/C)
     l1_inv = xw*(f0v0*a01+f0vt*(1/2/sw**2+a00))/2/C
     l2_inv = xw*f0v0*f1v0/2/C
     l3_inv = xw*f1v0/2/C/sw**2
     l4_inv = -xw*(2*a01/sw**2 - f0vt*f0v0)/4/C
     Poall = (np.pi)**3*(A*sw/N_inv)**2/np.abs(np.sqrt(Ld2_inv*Ls2_inv))*np.exp(-2*(xw/sw)**2)/C
     P1o1 = np.exp((xw/sw**2)**2/2/C)
     P1o2 = np.exp((1+2*a00*sw**2)*(xw/sw**2)**2/2/C)
     P1s1 = np.exp(l4_inv**2/Ld2_inv/2) + np.exp(-l3_inv**2/Ls2_inv/2)
     P1s2 = np.exp(-l1_inv**2/Ld2_inv/2) + np.exp(l2_inv**2/Ls2_inv/2)
     Purity1 = np.abs(Poall)*(P1o1*P1s1+P1o2*P1s2)
     P2e1 = (-l1_inv**2+l4_inv**2)/Ld2_inv + (l2_inv**2+l3_inv**2)/Ls2_inv
     P2e2 = 4*(1+a00)*(xw/sw)**2/C
     P2c = np.cos((l1_inv*l4_inv/Ld2_inv - l2_inv*l3_inv/Ls2_inv)/4)
     Purity2 = 8*Poall*np.exp((P2e1+P2e2)/8)*P2c
     # Renyi_2
     R2t = -np.log(Purity1+Purity2) 
     # Wigner decay              
     D2t = 10**3*A*np.sqrt(1/C/Ld2_inv)*np.exp((xw/2/sw**2)**2/C+l4_inv**2/Ld2_inv/4-(xw/sw)**2)  
     return R2t, D2t
    
    
     
                                                                                               
 
 # Main program starts here               
                                       

 # Producing G_R(t) from G_R(w)
Gw_vec = fGRw(full_w_vec)  # Discretizing G_R(w)
   # Inverse Fourier Transform of G_R(w), producing a time-series of G_R(t)
GR_t = np.real(ginvFT(Gw_vec, full_w_vec, full_t_ind, full_t_vec, full_t_ind)) 

 # Producing discretized G_sym(w)
G_sym = fG_sym(full_w_vec)
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     
                          
# solving elemetary function h(t) with h(0)=0, hdot(0)=1
rh = ode(h).set_integrator('vode', method='adams', with_jacobian=None)
rh.set_initial_value(hx0, t0).set_f_params(0.)
hxall=[]
hvall=[]
htall=[]
while rh.successful() and rh.t < t1:
#    print r.t
    rh.set_f_params(convoluh(rh.t,GR_t))
    rh.integrate(rh.t+dt)
#    print r.t, r.y
    htall.append(rh.t)    # time series
    hxall.append(rh.y[0]) # time series of h
    hvall.append(rh.y[1]) # time series of h'



# solving elemetary function H(t) with H(0)=1, Hdot(0)=0
rH = ode(H).set_integrator('vode', method='adams', with_jacobian=None)
rH.set_initial_value(Hx0, t0).set_f_params(0.)
Hxall=[]
Hvall=[]
Htall=[]
while rH.successful() and rH.t < t1:
#    print r.t
    rH.set_f_params(convoluh(rH.t,GR_t))
    rH.integrate(rH.t+dt)
#    print r.t, r.y
    Htall.append(rH.t)    # time series
    Hxall.append(rH.y[0]) # time series of H
    Hvall.append(rH.y[1]) # time series of H'


 
f1xall=[]
f1vall=[]
f0xall=[]
f0vall=[]
g0xall=[]
g1xall=[]
ht_all=[]
 


Renyi2=[]
Decayf=[]

t_ind = []
tf=tf_ind*dt
while tf <= t1:    
     tfin = int(tf/dt)
     t_ind = np.arange(tfin) # integer time index
     gw_vec = np.linspace(1/tf, 1/dt, tfin)   # frequency vector
     ht_all = np.array([htall[i] for i in t_ind])     # time vector
     f1xall = np.array([hxall[i]/hxall[tfin-1] for i in t_ind]) # time series of f1
     f1vall = np.array([hvall[i]/hxall[tfin-1] for i in t_ind])  # time series of f1'
     f0xall = np.array([Hxall[i] - f1xall[i]*Hxall[tfin-1] for i in t_ind])   # time series of f0
     f0vall = np.array([Hvall[i]-f1vall[i]*Hxall[tfin-1] for i in t_ind])     # time series of f0'
     f0v_0 = f0vall[0]       # f0v(0)
     f1v_0 = f1vall[0]       # f1v(0)
     f0v_t = f0vall[tfin-1]  # f0v(t)
     f1v_t = f1vall[tfin-1]  # f1v(t)
     g0xall = f1xall[::-1]   # g0(t)
     g1xall = f0xall[::-1]   # g1(t)
     g0_ftw = gFT(g0xall, ht_all, t_ind, gw_vec, t_ind)    #int_0^t g0(tau)*exp(-i.w.tau) dtau
     g1_ftw = gFT(g1xall, ht_all, t_ind, gw_vec, t_ind)    #int_0^t g1(tau)*exp(-i.w.tau) dtau
     a_sol = aij(g0_ftw, g1_ftw, gw_vec, G_sym, t_ind)   
     a_00, a_01, a_11 = a_sol[0], a_sol[1], a_sol[2]    
     WR_t = Wigner_Renyi(s_w,x_w,hxall[tfin-1],f0v_0,f1v_0,f0v_t,f1v_t,a_00,a_01,a_11)
     Renyi2.append(WR_t[0])
     Decayf.append(WR_t[1])
     tf += dt



t_vec_f = htall[tf_ind:]    # time vector for final time tf



                                                 


#htallo = TemporaryFile()
#np.save(htallo, htall)

#g0o = TemporaryFile()
#np.save(g0o, g0xall)

 

plt.plot(t_vec_f, Decayf)
plt.plot(t_vec_f, Renyi2)
plt.xlabel('t')
plt.ylabel('Wigner_Renyi2')

 
plt.show()
 
 